/*****************************************************************
 * grlib.h
 *
 * 1998.05.14  <tsakurai@rimath.saitama-u.ac.jp>
 *****************************************************************/

/********************
 * grlib*.c 
 -------------------*/
/* Screen Coordmode */
#define  Y_DOWN  0
#define  Y_UP    1
 
/* Mouse */
#define  LEFT_BUTTON   1  /* Button1 */
#define  MIDDLE_BUTTON 2  /* Button2 */
#define  RIGHT_BUTTON  3  /* Button3 */

#define  MOUSE_DOWN    0
#define  MOUSE_UP      1
#define  MOUSE_MOVE    2

#ifndef _Color_
typedef struct {
    unsigned char r, g, b;
    char id;
    unsigned long *cref; /* to save pixel-value or COLORREF */
} Color;
#define _Color_
#endif

#ifndef _Point2D_
typedef struct {
    double x, y;
} Point2D;
#define _Point2D_
#endif

/* functions defined in grlib*.c */
extern void InitCanvas(int width, int height, char *title);
extern void CloseCanvas(void);
extern void SetDoubleBuffer(int ON_OFF); 

extern int  KbHit(void);
extern int  GetKey(void);
extern void SetMouseFunc(void (*mouse)(int button, int state,
				       double x, double y));

extern void SetCoordMode(int Y_UP_DOWN);
extern void SetCoordOffset(double x, double y);
extern void ClearCanvas(void);
extern void FlushBuffer(void);

extern void SetColor(Color *c);
extern void SetColor3i(int r, int g, int b);
extern void SetColor3f(double r, double g, double b);
extern void SetGraystep(int step); 
extern void SetGray(double scale);
extern void SetColoredGray(Color *c, double scale);
extern void SetBackground(Color *c);
extern Color *GetBackground(void);

extern void DrawPoint(double x, double y);
extern void DrawPoints(Point2D *p, int N);
extern void SetLineWidth(double linewidth);
extern void DrawLine(double x1, double y1, double x2, double y2);
extern void DrawLines(Point2D *p, int N);
extern void DrawDashLine(double x1, double y1, double x2, double y2);
extern void FillRectangle(double x1, double y1, double x2, double y2);
extern void DrawRectangle(double x1, double y1, double x2, double y2);
extern void DrawRectangleXOR(double x1, double y1, double x2, double y2);
extern void FillPolygon(Point2D *p, int N);
extern void DrawPolygon(Point2D *p, int N);

extern void SelectFont(char *font_name);
extern void DrawString(double x, double y, char *str);

/********************
 * popup menu
 -------------------*/
typedef struct _MenuItem  MenuItem;
typedef struct _Menu      Menu;

struct _MenuItem {
    char  *label;
    int    id;
    Menu  *submenu;
    unsigned long win;    /* Window = XID = unsigned long */
};

struct _Menu {
    MenuItem *item;
    Menu     *parent;
    int   num_item;
    unsigned long win;    /* Window = XID = unsigned long */
};

extern void PopupMenu(Menu *menu, void (*menufunc)(int id));

/********************
 * save image
 -------------------*/
extern void SaveImageAs(char *fname);

/** END ***********************************************************/
