/******************************************************************
 *  grlibw.c: Graphic library for win32
 *            written  by Fumiya Horie (1994.07 C_Magazine)
 *            modified by <tsakurai@rimath.saitama-u.ac.jp> 1998.05
 ******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "grlib.h"

#define Y_UP   1
#define Y_DOWN 0

/* Mouse */
#define  LEFT_BUTTON   1  /* Button1 */
#define  MIDDLE_BUTTON 2  /* Button2 */
#define  RIGHT_BUTTON  3  /* Button3 */

#define  MOUSE_DOWN  0
#define  MOUSE_UP    1
#define  MOUSE_MOVE  2

#ifndef _Color_
typedef struct {
    unsigned char r, g, b;
    char id;
    unsigned long *cref; /* to save pixel-value */
} Color;
#define _Color_
#endif

#ifndef _Point2D_
typedef struct {
    double x, y;
} Point2D;
#define _Point2D_
#endif

/* Canvas configuration */
static int _width, _height;
static int _line_width = 0;
static int coordmode = Y_DOWN;
static int doublebuf = 0;   /* 1: draw back then copy to front */

static Color BGColor ={0,0,0};

static  Point2D O = {0,0};

#define IX(XX)  (int)(O.x+(XX)+.1)
#define IY(YY)  ((coordmode==Y_UP) ? (int)(_height-1-O.y-(YY)+.1)\
		                   : (int)(O.y+(YY)+.1))

#define XX(IX)  ((IX)-O.x)
#define YY(IY)  ((coordmode==Y_UP) ? (_height-1-(IY)-O.y) : ((IY)-O.y))

#define TEXTHEIGHT 13
#define TEXTWIDTH   7

/****************************************************************/

static HINSTANCE hinst;    /* instance */
static HWND win;           /* main window */
static HDC  canvas;        /* fore or back */
static HDC  hdc;           /* foreground DC */
static HDC  hdc_back;      /* background DC */
static HBITMAP bmp_back;   /* background BMP */
static HFONT  hfont;
static COLORREF current_rgb = RGB(255,255,255); /* COLORREF = ulong */
static HPEN   current_pen   = 0;
static HBRUSH current_brush = 0;
/****************************************************************/

static LRESULT CALLBACK window_proc(HWND hWnd, UINT wMsg,
				    WPARAM wParam, LPARAM lParam)
{
    switch (wMsg){
    case WM_PAINT:
	BitBlt(hdc, 0, 0, _width, _height, hdc_back, 0, 0, SRCCOPY);
	return DefWindowProc(hWnd, wMsg, wParam, lParam);
    case WM_CLOSE:
	DestroyWindow(hWnd);
	PostQuitMessage(0);
	exit(0);
    default:
	return DefWindowProc(hWnd, wMsg, wParam, lParam);
    }
}

void InitCanvas(int width,int height, char *title)
{
    int w, h;
    RECT rect;
    WNDCLASS wclass;
    DWORD wstyle = (0)? WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX
	              : WS_OVERLAPPEDWINDOW;
	       
    wclass.lpszClassName ="Display";
    wclass.lpszMenuName  = NULL;
    wclass.hInstance     = GetModuleHandle(NULL);
    wclass.lpfnWndProc   = window_proc;
    wclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wclass.hIcon         = NULL;
    wclass.hbrBackground = GetStockObject(BLACK_BRUSH);
    wclass.style         = 0;
    wclass.cbClsExtra    = 0;
    wclass.cbWndExtra    = 0;

    if (!RegisterClass(&wclass))  exit(1);

    _width = width; _height= height;
    SetRect(&rect, 0, 0, width, height);
    AdjustWindowRect(&rect, wstyle, 0);
    w = rect.right  - rect.left;
    h = rect.bottom - rect.top;

    win = CreateWindow(wclass.lpszClassName, title, wstyle,
		       CW_USEDEFAULT, CW_USEDEFAULT,
		       w, h, NULL, NULL, wclass.hInstance, NULL);
    ShowWindow(win, SW_SHOWDEFAULT);

    /* set as default */
    canvas = hdc = GetDC(win);
    coordmode = Y_DOWN;

    /* set up background bitmap */
    hdc_back = CreateCompatibleDC(hdc);
    bmp_back = CreateCompatibleBitmap(hdc, width, height);
    SelectObject(hdc_back, bmp_back);

    /* font */
    hfont = CreateFont(TEXTHEIGHT, TEXTWIDTH, 0, 0, 
		       0, FALSE, FALSE, FALSE, DEFAULT_CHARSET,
		       OUT_DEFAULT_PRECIS,CLIP_TT_ALWAYS,PROOF_QUALITY,
		       FIXED_PITCH|FF_ROMAN, "TERMINAL");
    SelectObject(hdc, hfont);
    SelectObject(hdc_back, hfont);

    FlushConsoleInputBuffer(GetStdHandle(STD_INPUT_HANDLE));
}

void CloseCanvas(void)
{
    if (win) {
	DeleteDC(hdc);
	DeleteDC(hdc_back);
	DeleteObject(bmp_back);
	if (current_pen) DeleteObject(current_pen);
	if (current_brush) DeleteObject(current_brush);
	UnregisterClass("Display", GetModuleHandle(NULL));
	win = 0;
    }
}

void SetDoubleBuffer(int on) 
{
    if (on) {
	canvas = hdc_back;
	doublebuf = 1;
    } else {
	canvas = hdc;
	doublebuf = 0;
    }
}

/*************************************
 * Save Image
 ------------------------------------*/
#define BM        (19778)   /* = ('B' | ('M'<<8)) */
#define DEPTH     (24)
#define bfOffBits (14+40)

static void put_header_bmp24(FILE *fp, int rx, int ry)
{
    BITMAPFILEHEADER  bfh = {BM, bfOffBits + 640*480, 0, 0, bfOffBits};
    BITMAPINFOHEADER  bih = {40, 640, 480, 1, DEPTH, 0, 0, 0, 0, 0, 0};

    bfh.bfSize = bfOffBits + (unsigned long)rx*ry;
    bih.biWidth  = rx;
    bih.biHeight = ry;
    fwrite(&bfh, 14, 1, fp);
    fwrite(&bih, 40, 1, fp);
}

static void save_bmp24(FILE *fp, int x1, int y1, int x2, int y2)
{
    int x, y;
    
    put_header_bmp24(fp, x2-x1, y2-y1);
    for(y=y2-1; y>=y1; y--) {
	for (x=x1; x<x2; x++) {
	    COLORREF cref = GetPixel(hdc_back, x, y);
	    putc(GetBValue(cref), fp);
	    putc(GetGValue(cref), fp);
	    putc(GetRValue(cref), fp);
	}
        while ((x-x1)%4 != 0) { putc('\0', fp); x++; }
    }
}

void SaveImageAs(char *fname)
{
    FILE *fp;
    char str[100];

    if ((fp=fopen(fname, "wb")) == 0) {
	fprintf(stderr, "can't open %s\n", fname);
	return;
    }
    SetColor3f(0,1,0);
    sprintf(str, "Saving Image to \"%s\" ", fname);
    TextOut(hdc, _width/2-50, _height/2-14, str, strlen(str));
    save_bmp24(fp, 0, 0, _width, _height);
    TextOut(hdc, _width/2-50, _height/2-14+20, "Done!", 5);
    fclose(fp);
}

/*************************************
 * popup-menu
 ------------------------------------*/
static void (*menuCB)(int id);

static unsigned int setupMenu(Menu *menu)
{
    HMENU hmenu;
    MenuItem *item = menu->item;

    hmenu = CreatePopupMenu();
    while (item->label) {
	if (item->submenu) {
	    unsigned hsubmenu = setupMenu(item->submenu);
	    AppendMenu(hmenu, MF_POPUP | MF_STRING, 
		       (unsigned)hsubmenu, item->label);
	} else
	    AppendMenu(hmenu, MF_STRING, item->id, item->label);
	item++;
    }
    return (unsigned)hmenu;
}

#ifndef TPM_RETURNCMD   /* not defined in Cygnus header */
#define TPM_RETURNCMD 0x0100l
#endif

void PopupMenu(Menu *menu, void (*menufunc)(int id))
{
    static unsigned int menu_id = 0;
    int select;
    POINT pt;
    MSG  msg;

    if (menu_id == 0){
	menuCB  = menufunc;
	menu_id = setupMenu(menu);
    }
    GetCursorPos(&pt);
    select = TrackPopupMenu((HMENU) menu_id, 
			    TPM_LEFTALIGN|TPM_RIGHTBUTTON|TPM_RETURNCMD,
			    pt.x, pt.y, 0, win, NULL);
    if (select)
	menuCB(select);
    else 
	GetMessage(&msg, win, 0, 0);
}

/*************************************
 * Mouse event
 ------------------------------------*/
static void (*mouseCB)(int button, int state, double x, double y) = NULL;
    
void SetMouseFunc(void (*mouse)(int, int, double, double)) { mouseCB = mouse; }

static void mouseEvent(MSG msg)
{
    static int lftdwn=0;
    POINT pt;
    
    GetCursorPos(&pt);
    ScreenToClient(win, &pt);
    switch (msg.message) {
    case WM_RBUTTONDOWN:
	mouseCB(RIGHT_BUTTON, MOUSE_DOWN, XX(pt.x), YY(pt.y));
	return;
    case WM_RBUTTONUP:
	mouseCB(RIGHT_BUTTON, MOUSE_UP, XX(pt.x), YY(pt.y));
	return;
    case WM_LBUTTONDOWN:
	mouseCB(LEFT_BUTTON, MOUSE_DOWN, XX(pt.x), YY(pt.y));
	lftdwn=1; return;
    case WM_LBUTTONUP:
	mouseCB(LEFT_BUTTON, MOUSE_UP, XX(pt.x), YY(pt.y));
	lftdwn=0; return;
    case WM_MOUSEMOVE:
	if (lftdwn == 1)
	    mouseCB(LEFT_BUTTON, MOUSE_MOVE, XX(pt.x), YY(pt.y));
	return;
    default:
	return;
    }
}

/*************************************
 * Key event
 ------------------------------------*/
static int last_key = -1;

int KbHit(void)
{
    MSG msg;
    while (PeekMessage(&msg,0,0,0,PM_REMOVE)) {
	TranslateMessage(&msg);
	if (mouseCB) {
	    mouseEvent(msg);
	}
	if (msg.message == WM_CHAR) {
	    last_key = msg.wParam;
	    return 1;
	} else
	    DispatchMessage(&msg);
    }
    return 0;
}

int GetKey(void)
{
    int k;

    while (last_key == -1)
	KbHit();
    k = last_key;
    last_key = -1;
    return k;
}

/******************************
 * common functions
 -----------------------------*/
void SetCoordMode(int mode)
{
    coordmode = mode;  /* mode = Y_UP / Y_DOWN  */ 
}

void SetCoordOffset(double x0, double y0)
{
    O.x = x0;
    O.y = y0;
}

void FlushBuffer(void)
{
    GdiFlush();
    if (doublebuf)
	BitBlt(hdc, 0, 0, _width, _height, hdc_back, 0, 0, SRCCOPY);
    else 
	BitBlt(hdc_back, 0, 0, _width, _height, hdc, 0, 0, SRCCOPY);
}

void ClearCanvas(void)
{
    HBRUSH new_brush;

    SetColor(&BGColor);
    new_brush = CreateSolidBrush(current_rgb);
    SelectObject(canvas, new_brush);
    if (current_brush) DeleteObject(current_brush);
    current_brush = new_brush;
    Rectangle(canvas, -10, -10, _width+10, _height+10);
}

/*************************************
 * Set Color
 ------------------------------------*/
void  SetBackground(Color *c) { BGColor = *c; }
Color *GetBackground(void)    { return &BGColor; }

void SetColor(Color *c) { current_rgb = RGB(c->r,c->g,c->b); }

void SetColor3i(int r, int g, int b)
{
    current_rgb = RGB(r,g,b);
}

void SetColor3f(double r, double g, double b)
{
    Color c;
    c.r = (int)(255*r); c.g = (int)(255*g); c.b = (int)(255*b);
    current_rgb = RGB(c.r,c.g,c.b);
}

static int wgraystep; /* not used */

void SetGraystep(int N)
{
    wgraystep = N;    /* have no meaning */
}

void SetGray(double scale)
{
    Color c;
    if (scale > 1) scale = 1;
    if (scale <-1) scale =-1;
    if (scale >= 0)
	c.r = c.g = c.b = (int) (scale*255);
    else {
	c.r = (int)(-scale*255); 
	c.g = c.b = (int)(-scale*128); 
    }
    current_rgb = RGB(c.r,c.g,c.b);
}

void SetColoredGray(Color *color, double scale)
{
    Color c;

    if (scale > 1) scale = 1;
    if (scale <-1) scale =-1;
    if (scale >= 0 ){
	c.r = (int)(color->r*scale);
	c.g = (int)(color->g*scale);
	c.b = (int)(color->b*scale);
    } else {
	c.r = (int)(-color->r*scale);
	c.g = (int)(-color->g/2*scale);
	c.b = (int)(-color->b/2*scale);
    }
    current_rgb = RGB(c.r,c.g,c.b);
}

/*************************************
 * Drawing functions
 ------------------------------------*/
static void newPen(void)
{
    static COLORREF pen_rgb;
    HPEN new_pen;
    
    if (pen_rgb == current_rgb) return;
    new_pen = CreatePen(PS_SOLID, _line_width, current_rgb);
    SelectObject(canvas, new_pen);
    if (current_pen) DeleteObject(current_pen);
    current_pen = new_pen;
    pen_rgb = current_rgb;
}

static void newBrush(void)
{
    static COLORREF brush_rgb = RGB(0,0,0);
    HBRUSH new_brush;

    if (brush_rgb == current_rgb) return;
    new_brush = CreateSolidBrush(current_rgb);
    SelectObject(canvas, new_brush);
    if (current_brush) DeleteObject(current_brush);
    current_brush = new_brush;
    brush_rgb = current_rgb;
}

void DrawPoint(double x, double y)
{
    SetPixel(canvas, IX(x), IY(y), current_rgb);
}

void DrawPoints(Point2D *p, int N)
{
    for (; N--; p++)
	SetPixel(canvas, IX(p->x), IY(p->y), current_rgb);
}

void SetLineWidth(double line_width)
{
    if (_line_width == (int)line_width) return;
    _line_width = (int)line_width;
    newPen();
}

void DrawLine(double x0, double y0, double x1, double y1)
{
    newPen();
    MoveToEx(canvas, IX(x0), IY(y0), NULL);
    LineTo(canvas, IX(x1), IY(y1));
    /* SetPixel(canvas, IX(x1), IY(y1), current_rgb); */
}

void DrawLines(Point2D *p, int N)
{
    newPen();
    MoveToEx(canvas, IX(p->x), IY(p->y), NULL);
    while (--N) {
	p++;
	LineTo(canvas, IX(p->x), IY(p->y));
    }
    /* SetPixel(canvas, IX(p->x), IY(p->y), current_rgb); */
}

void DrawDashLine(double x0, double y0, double x1, double y1)
{
    HPEN new_pen, old_pen;

    new_pen = CreatePen(PS_DOT, 0, current_rgb);
    old_pen = SelectObject(canvas, new_pen);

    MoveToEx(canvas, IX(x0), IY(y0), NULL);
    LineTo(canvas, IX(x1), IY(y1));
    SetPixel(canvas,IX(x1),IY(y1),current_rgb);

    SelectObject(canvas, old_pen);
    DeleteObject(new_pen);
}

void FillRectangle(double x1, double y1, double x2, double y2)
{
    RECT rect;
    int 
        ix1 = IX(x1), iy1 = IY(y1),
        ix2 = IX(x2), iy2 = IY(y2),
        swap;
    if (ix1 > ix2) { swap = ix1; ix1 = ix2; ix2 = swap; } 
    if (iy1 > iy2) { swap = iy1; iy1 = iy2; iy2 = swap; } 
    newBrush(); newPen();
    SetRect(&rect, ix1, iy1, ix2, iy2);
    FillRect(canvas, &rect, current_brush);
}

void DrawRectangle(double x1, double y1, double x2, double y2)
{
    newPen();
    MoveToEx(canvas, IX(x1), IY(y1), NULL);
    LineTo(canvas, IX(x1), IY(y2));
    LineTo(canvas, IX(x2), IY(y2));
    LineTo(canvas, IX(x2), IY(y1));
    LineTo(canvas, IX(x1), IY(y1));
}

void DrawRectangleXOR(double x1, double y1, double x2, double y2)
{
    HPEN 
        new_pen = CreatePen(PS_SOLID, 0, RGB(255,255,255)),
        old_pen = SelectObject(hdc, new_pen);

    SetROP2(hdc, R2_XORPEN);
    MoveToEx(hdc, IX(x1), IY(y1), NULL);
    LineTo(hdc, IX(x1), IY(y2));
    LineTo(hdc, IX(x2), IY(y2));
    LineTo(hdc, IX(x2), IY(y1));
    LineTo(hdc, IX(x1), IY(y1));

    SetROP2(hdc, R2_COPYPEN);
    SelectObject(hdc, old_pen);
    DeleteObject(new_pen);
}

void FillPolygon(Point2D *p, int N)
{
    int i;
    POINT *pt=(POINT *)malloc(N*sizeof(POINT));

    for (i=0; i<N; i++) {
	pt[i].x = IX(p[i].x);  pt[i].y = IY(p[i].y);
    }
    Polygon(canvas, pt, N);
    free(pt);
}

void DrawPolygon(Point2D *p, int N)
{
    MoveToEx(canvas, IX(p[N-1].x), IY(p[N-1].y), NULL);
    for (; N--; p++)
	LineTo(canvas, IX(p->x), IY(p->y));
}

void SelectFont(char *fontname)
{
    hfont = CreateFont(TEXTHEIGHT, TEXTWIDTH, 0, 0, 
		       0, FALSE, FALSE, FALSE, DEFAULT_CHARSET,
		       OUT_DEFAULT_PRECIS,CLIP_TT_ALWAYS,PROOF_QUALITY,
		       FIXED_PITCH|FF_ROMAN, fontname);
    SelectObject(canvas, hfont);
}

void DrawString(double x, double y, char *str)
{
    SetTextColor(canvas, current_rgb);
    SetBkMode(canvas, TRANSPARENT);
    TextOut(canvas, IX(x), IY(y)-TEXTHEIGHT, str, strlen(str));
}

/*********************************************/
#undef IX
#undef IY
#undef XX
#undef YY

/* End */
