/*********************************************************************
 * grlibx.c: Graphic library for X
 *           written  by Ķɴ (1994.07 C-Magazine)
 *           modified by <tsakurai@rimath.saitama-u.ac.jp> 1998.05.26
 *********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "grlib.h"

/* Canvas Coordmode */
#define Y_DOWN 0
#define Y_UP   1
 
/* Mouse */
#define  LEFT_BUTTON   1  /* Button1 */
#define  MIDDLE_BUTTON 2  /* Button2 */
#define  RIGHT_BUTTON  3  /* Button3 */

#define  MOUSE_DOWN   0
#define  MOUSE_UP     1
#define  MOUSE_MOVE   2

#define  USE_POPUP_MENU 

/* structs */
#ifndef _Color_
typedef struct {
    unsigned char r, g, b;
    char id;
    unsigned long *cref; /* to save pixel-value */
} Color;
#define _Color_
#endif

#ifndef _Point2D_
typedef struct {
    double x, y;
} Point2D;
#define _Point2D_
#endif

/********************************************************************/
/* Canvas configuration */
static int  xwidth, xheight;
static int  coordmode = 0;    /* Y_DOWN, 1: Y_UP */ 
static int  doublebuffer = 0; /* 1: draw back then copy to front */

static Color BGColor = {0, 0, 0};

/* default graystep for 8bpp-server */
static int xgraystep = 128;

/* set up screen coordinate */
static Point2D  O = {0.0, 0.0};

#define IX(XX)  ((int)((XX) + O.x +.1))
#define IY(YY)  ((coordmode == Y_UP) ? (int)(xheight-1 - (YY) - O.y + .1) \
		                     : (int)((YY) + O.y + .1))
#define XX(IX)  ((IX) - O.x)
#define YY(IY)  ((coordmode==Y_UP) ? (xheight-1 - (IY) - O.y) : ((IY) - O.y))

static Display *disp;
static Window   win;
static Pixmap   pmap;
static Drawable canvas;
static Colormap cmap;
static GC       gc;
static XEvent   event;
static XFontStruct *xfont;
static char default_font[] = "fixed";

static unsigned long  xblack, xwhite;
static unsigned long  *xgray;
static unsigned long  *xcgray[32];
static int  xcgraystep[32];

#define DoRGB       ( DoRed | DoGreen | DoBlue )
#define Event_Mask  ( ExposureMask | KeyPressMask | Button1MotionMask | \
		      ButtonPressMask | ButtonReleaseMask )

#define XSetBackingStore(DISP, WIN) { \
           XSetWindowAttributes wa;   \
           wa.backing_store = Always; \
  	   XChangeWindowAttributes(DISP, WIN, CWBackingStore, &wa);}

#define XSetOverrideRedirect(DISP, WIN) { \
           XSetWindowAttributes wa;       \
           wa.override_redirect = True;   \
           XChangeWindowAttributes(DISP, WIN, CWOverrideRedirect, &wa);}

#define XSetLineWidth(DISP, GC, WIDTH)  XSetLineAttributes( \
                      DISP, GC, WIDTH, LineSolid, CapButt, JoinMiter)

/******************************
 * Init Display device 
 -----------------------------*/
void InitCanvas(int width, int height, char *title)
{
    if ((disp = XOpenDisplay(NULL)) == NULL) {
	fputs("cannot open display.\n", stderr);
	exit(1);
    }
    xwidth = width;
    xheight= height;
    xblack = BlackPixel(disp, 0);
    xwhite = WhitePixel(disp, 0);
    xfont  = XLoadQueryFont(disp, default_font);
    cmap   = DefaultColormap(disp, 0);
    gc     = XCreateGC(disp, RootWindow(disp, 0), 0, 0);
    win    = XCreateSimpleWindow(disp, RootWindow(disp, 0),
				 10, 10, xwidth, xheight, 1, xwhite, xblack);
    pmap   = XCreatePixmap(disp, win, xwidth, xheight, DefaultDepth(disp, 0));
    XMapWindow(disp, win);
    XStoreName(disp, win, title);
    XSelectInput(disp, win, Event_Mask);
    XSetBackingStore(disp, win);
    canvas = win;
}

void CloseCanvas(void)
{
    int i;
    
    if (xgray) free(xgray);
    for (i = 1; xcgray[i]; i++)  free(xcgray[i]);
    XFreeGC(disp, gc);
    XCloseDisplay(disp);
}

void SetDoubleBuffer(int ON) 
{
    if (ON) {
	canvas = pmap;
	doublebuffer = 1;
    } else {
	canvas = win;
	doublebuffer = 0;
    }
}

/******************************
 * save image
 -----------------------------*/
void SaveImageAs(char *fname)
{
    fprintf(stderr, "\nNot implemented.\n");
    fprintf(stderr, "Use \"xwd > file-name\" instead.\n");
    fflush(stderr);
}

/******************************
 * popup menu
 -----------------------------*/
#ifdef USE_POPUP_MENU
#include "smenux.c"
#endif

/******************************
 * Mouse event
 -----------------------------*/
static void (*mouseCB)(int button, int state, double x, double y) = NULL;

void SetMouseFunc(void (*mouse)(int button, int state, double x, double y))
{
    mouseCB = mouse;
}

/******************************
 * Key event
 -----------------------------*/
int KbHit(void)
{
    if (!XCheckMaskEvent(disp, Event_Mask, &event))
	return 0;
    switch (event.type){
    case Expose:
	XFlush(disp);
	return 0;
    /* mouse event */
    case ButtonPress:
	if (mouseCB)
	    mouseCB(event.xbutton.button, MOUSE_DOWN, 
			   XX(event.xbutton.x), YY(event.xbutton.y));
	return 0;
    case ButtonRelease:
	if (mouseCB)
	    mouseCB(event.xbutton.button, MOUSE_UP, 
		    XX(event.xbutton.x), YY(event.xbutton.y));
	return 0;
    case MotionNotify:
	if (mouseCB)
	    mouseCB(LEFT_BUTTON, MOUSE_MOVE,
		    XX(event.xbutton.x), YY(event.xbutton.y));
	return 0;
    default:
	return (event.type == KeyPress);
    }
}

int GetKey(void)
{
    char str[3];
    
    while (event.type != KeyPress)
	KbHit();
    if (XLookupString(&event.xkey, str, sizeof(str), (KeySym *)NULL,
		      (XComposeStatus *)NULL) > 0)
	return str[0];
    else
	return 0;
}

/******************************
 * common func
 -----------------------------*/
void SetCoordMode(int mode)
{
    coordmode = mode;  /* mode: Y_UP =1, Y_DOWN =0 */ 
}

void SetCoordOffset(double x, double y)
{
    O.x = x, O.y = y;
}

void ClearCanvas(void)
{
    SetColor(&BGColor);
    XFillRectangle(disp, canvas, gc, 0, 0, xwidth, xheight);
}

void FlushBuffer(void)
{
    if (doublebuffer)
	XCopyArea(disp, pmap, win, gc, 0, 0, xwidth, xheight, 0, 0);
    XFlush(disp);
}

/*************************************
 * Set Color
 ------------------------------------*/
void SetBackground(Color *c)
{
    BGColor = *c;
}

Color *GetBackground(void)
{
    return &BGColor;
}

static unsigned long get_xpixel(XColor *xc)
{
    if (!XAllocColor(disp, cmap, xc)) {
	fprintf(stderr, "can't alloc color {%d,%d,%d}\n",
		xc->red/256, xc->green/256, xc->blue/256);
	fflush(stderr);
	return BlackPixel(disp, 0);
    } else 
	return xc->pixel;
}

void SetColor(Color *c)
{
    if (c->cref)
	XSetForeground(disp, gc, *(c->cref));
    else {
	XColor xc;
	xc.red = c->r*256, xc.green = c->g*256, xc.blue = c->b*256;
	c->cref = (unsigned long *)malloc(sizeof(unsigned long));
	*(c->cref) = get_xpixel(&xc);
	XSetForeground(disp, gc, *(c->cref));
    }
}

void SetColor3i(int r, int g, int b)
{
    XColor xc;
    xc.red   = r*256;
    xc.green = g*256;
    xc.blue  = b*256;
    XSetForeground(disp, gc, get_xpixel(&xc));
}

void SetColor3f(double r, double g, double b)
{
    XColor xc;
    xc.red   = (int)(r*65535);
    xc.green = (int)(g*65535);
    xc.blue  = (int)(b*65535);
    XSetForeground(disp, gc, get_xpixel(&xc));
}

void SetGraystep(int step) 
{
    xgraystep = step;
    printf("graystep = %d\n", xgraystep);
}

static void alloc_gray(int N)
{
    XColor xc;
    int i;

    if (xgray == NULL)
	xgray = (unsigned long *)malloc(N*sizeof(unsigned long));
    else 
	xgray = (unsigned long *)realloc(xgray, N*sizeof(unsigned long));
    for (i = 0; i < N; i++) {
	xc.red = xc.green = xc.blue = 65536*i/N;
	xgray[i] = get_xpixel(&xc);
    }
}

void SetGray(double scale)
{
    int i;
    
    if (scale == 0)
	XSetForeground(disp, gc, xblack);
    else  if (scale == 1)
	XSetForeground(disp, gc, xwhite);
    else {
	if (xgray == NULL) alloc_gray(xgraystep);
	if (scale > 1.0) scale = 1.0;
	if (scale <-1.0) scale =-1.0;
	i = (int)((xgraystep-1)*scale);
	i = (i >= 0) ? i : -i;
	XSetForeground(disp, gc, xgray[i]);
    }
}

static void alloc_cgray(Color *c, int N)
{
    static int id = 0;
    int i;
    XColor xc, xc_back;
    
    c->id = ++id;
    xcgraystep[id] = N;
    if (xcgray[id] == NULL)
	xcgray[id] = (unsigned long *)malloc(2*N*sizeof(unsigned long));
    else
        xcgray[id] = (unsigned long *)realloc(xcgray[id], 
					   2*N*sizeof(unsigned long));
    for (i = N-1; i >= 0; i--){
	xc.red   = (c->r)*256*i/N;
	xc.green = (c->g)*256*i/N;
	xc.blue  = (c->b)*256*i/N;
	xcgray[id][N+i] = get_xpixel(&xc);
    }
    if (id == 1) {                     /* ΢̤ 1 Τ */
	for (i=(N-1)/2; i>=0; i--){    /* ΢̿ graystep  N/2 */
	    xc_back.red   = (c->r)  *256*2*i/N;
	    xc_back.green = (c->g/2)*256*2*i/N;
	    xc_back.blue  = (c->b/2)*256*2*i/N;
	    xcgray[id][N-2*i] = xcgray[id][N-2*i-1] = get_xpixel(&xc_back);
	}
    } else {
	for (i = 0; i < N; i++)
	    xcgray[id][N-i] = xcgray[1][xcgraystep[1]*(N-i)/N];
    }
}

void SetColoredGray(Color *c, double scale)
{
    int i, N;
    
    if (c->id == 0)
	alloc_cgray(c, xgraystep);
    N = xcgraystep[(int)c->id];
    if (scale >  1.0) scale = 1.0;
    if (scale < -1.0) scale =-1.0;
    i = (int)((N-1)*scale);
    XSetForeground(disp, gc, xcgray[(int)c->id][N+i]);
}

/*************************************
 * Drawing functions
 ------------------------------------*/
void DrawPoint(double x, double y)
{
    XDrawPoint(disp, canvas, gc, IX(x), IY(y));
}

void DrawPoints(Point2D *p, int N)
{
    for (; N--; p++)
	DrawPoint(p->x, p->y);
}

void SetLineWidth(double linewidth)
{
    XSetLineWidth(disp, gc, (int)linewidth);
}

void DrawLine(double x1, double y1, double x2, double y2)
{
    XDrawLine(disp, canvas, gc, IX(x1), IY(y1), IX(x2), IY(y2));
}

void DrawLines(Point2D *p, int N)
{
    for (; --N; p++)
	XDrawLine(disp, canvas, gc,
		  IX(p->x), IY(p->y), IX((p+1)->x), IY((p+1)->y));
}

void DrawDashLine(double x1, double y1, double x2, double y2)
{
    XSetLineAttributes(disp, gc, 0, LineOnOffDash, CapButt, JoinMiter);
    XDrawLine(disp, canvas, gc, IX(x1), IY(y1), IX(x2), IY(y2));
    XSetLineAttributes(disp, gc, 0, LineSolid, CapButt, JoinMiter);
}

void FillRectangle(double x1, double y1, double x2, double y2)
{
    int 
        ix1 = IX(x1), iy1 = IY(y1),
        ix2 = IX(x2), iy2 = IY(y2);
    int swap;
    if (ix1 > ix2) { swap = ix1; ix1 = ix2; ix2 = swap; } 
    if (iy1 > iy2) { swap = iy1; iy1 = iy2; iy2 = swap; } 
    XFillRectangle(disp, canvas, gc, ix1, iy1, ix2 - ix1, iy2 - iy1);
}

void DrawRectangle(double x1, double y1, double x2, double y2)
{
    int
        ix1 = IX(x1), iy1 = IY(y1),
        ix2 = IX(x2), iy2 = IY(y2);
    int swap;
    if (ix1 > ix2) { swap = ix1; ix1 = ix2; ix2 = swap; } 
    if (iy1 > iy2) { swap = iy1; iy1 = iy2; iy2 = swap; } 
    XDrawRectangle(disp, canvas, gc, ix1, iy1, ix2 - ix1, iy2 - iy1);
}

void DrawRectangleXOR(double x1, double y1, double x2, double y2)
{
    XSetFunction(disp, gc, GXxor);
    XSetForeground(disp, gc, 0xffffffff);
    DrawRectangle(x1, y1, x2, y2);
    XSetFunction(disp, gc, GXcopy);
}

void FillPolygon(Point2D *p, int N)
{
    int i;
    XPoint *xp = (XPoint *)malloc(N*sizeof(XPoint));
    for (i=0; i<N; i++) {
	xp[i].x = IX(p[i].x);  xp[i].y = IY(p[i].y);
    }
    XFillPolygon(disp, canvas, gc, xp, N, Nonconvex, CoordModeOrigin);
    free(xp);
}

void DrawPolygon(Point2D *p, int N)
{
    Point2D *p0=p;
    for (; --N; p++)
	XDrawLine(disp,canvas,gc, IX(p->x),IY(p->y),IX((p+1)->x),IY((p+1)->y));
    XDrawLine(disp, canvas, gc, IX(p->x), IY(p->y), IX(p0->x), IY(p0->y));
}

void SelectFont(char *font_name)
{
    xfont = XLoadQueryFont(disp, font_name);
}

void DrawString(double x, double y, char *str)
{
    XSetFont(disp, gc, xfont->fid);
    XDrawString(disp, canvas, gc, IX(x), IY(y), str, strlen(str));
}

/**********************************************/
#undef IX
#undef IY
#undef YX
#undef YY

/* END */
