/*****************************************************************
 * smenux.c: simple pop-up menu for X
 *
 * 1998.05.12  <tsakurai@rimath.saitama-u.ac.jp>
 *****************************************************************/ 

#include <X11/cursorfont.h> 

/* common variable with grlibx.c */
extern Display  *disp;    /* initialized in InitScreen() */

#define rwin  RootWindow(disp, 0)
#define cmap  DefaultColormap(disp, 0)

/* constants */
#define MENU_BORDER      0
#define ITEM_BORDER      0
#define ITEM_HEIGHT     20
#define MIN_ITEM_WIDTH  90
#define MAX_ITEM_WIDTH 200

/* color definition */
static Color dimgray   = {105, 105, 105};
static Color gray      = {180, 180, 180};
static Color lightgray = {211, 211, 211};

#define menuDimGray     GetPixel(&dimgray)
#define menuGray        GetPixel(&gray)
#define menuLightGray   GetPixel(&lightgray)
#define menuBlack       BlackPixel(disp, 0)
#define menuWhite       WhitePixel(disp, 0)

/* resources */
static Cursor  menuCursor;
static GC      menuGC;
static XFontStruct *menuFont;
static char    menufontname[] = 
                   "-adobe-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*";

/* dummy variables */
static int     JunkX, JunkY;
static Window  JunkRoot, JunkChild;
static unsigned int JunkWidth, JunkHeight, JunkBW, JunkDepth, JunkMask;

/* macros */
#define XSetBackingStore(DISP, WIN) {\
		XSetWindowAttributes wa; wa.backing_store = Always;\
		XChangeWindowAttributes(DISP, WIN, CWBackingStore, &wa);}
#define XSetOverrideRedirect(DISP, WIN) {\
		XSetWindowAttributes wa; wa.override_redirect = True;\
		XChangeWindowAttributes(DISP, WIN, CWOverrideRedirect, &wa);}
#define XGetMousePos(DISP, WIN, PX, PY)  \
                XQueryPointer(DISP, WIN, &JunkRoot, &JunkChild, \
			      &JunkX, &JunkY, PX, PY, &JunkMask)
#define XGetWindowSize(DISP, WIN, PW, PH)  \
                XGetGeometry(DISP, WIN, &JunkRoot, &JunkX, &JunkY, \
			     PW, PH, &JunkBW, &JunkDepth)

/****************************************************************/

static unsigned long GetPixel(Color *c)
{
    if (c->cref)
	return *(c->cref);
    else {
	XColor xc;
	xc.red = c->r*256, xc.green = c->g*256, xc.blue = c->b*256;
	c->cref = (unsigned long *)malloc(sizeof(unsigned long));
	if (!XAllocColor(disp, cmap, &xc)) {
	    fprintf(stderr, "can't alloc color {%d,%d,%d}\n", c->r,c->g,c->b);
	    return (*(c->cref) = WhitePixel(disp, 0));
	} else 
	    return (*(c->cref) = xc.pixel);
    }
}

static void InitMenu(Menu *menu)
{
    unsigned int width = MIN_ITEM_WIDTH;
    unsigned int height;
    int i;

    menu->win =	XCreateSimpleWindow(disp, rwin, 0, 0, 1, 1, MENU_BORDER,
				    menuBlack, menuWhite);
    XSetOverrideRedirect(disp, menu->win);
    XSetBackingStore(disp, menu->win);
    XSelectInput(disp, menu->win, PointerMotionMask);
    menuGC = XCreateGC(disp, menu->win, 0, 0);
    menuCursor = XCreateFontCursor(disp, XC_arrow);
    menuFont = XLoadQueryFont(disp, menufontname);
    XSetFont(disp, menuGC, menuFont->fid);
    for (i=0; menu->item[i].label; i++) {
	unsigned int itemwidth;
	char *label = menu->item[i].label;
	if (menu->item[i].submenu) {
	    menu->item[i].submenu->parent = menu;
	    InitMenu(menu->item[i].submenu);
	}
	menu->item[i].win = XCreateSimpleWindow ( 
				disp, menu->win, 0, ITEM_HEIGHT*(i),
				MIN_ITEM_WIDTH, ITEM_HEIGHT, ITEM_BORDER,
				menuBlack, menuWhite );
	XSelectInput(disp, menu->item[i].win, EnterWindowMask|LeaveWindowMask);
	XSetBackingStore(disp, menu->item[i].win);
	XMapWindow(disp, menu->item[i].win);
	itemwidth = XTextWidth(menuFont, label, strlen(label)) + 17;
	itemwidth += (menu->item[i].submenu) ? 20 : 0;
	width = (itemwidth > width)? itemwidth : width;
    }
    width = (width > MAX_ITEM_WIDTH)? MAX_ITEM_WIDTH : width;
    height = ITEM_HEIGHT*(menu->num_item=i);
    XResizeWindow(disp, menu->win, width, height);
    for (i=0; menu->item[i].label; i++)
	XResizeWindow(disp, menu->item[i].win, width, ITEM_HEIGHT);
}

static void RelieveRectangle(Window win, int x, int y, int w, int h)
{
    XSetForeground(disp, menuGC, menuWhite);
    XFillRectangle(disp, win, menuGC, x, y, w-1, 1);
    XFillRectangle(disp, win, menuGC, x, y, 1, h-1);

    XSetForeground(disp, menuGC, menuDimGray);
    XFillRectangle(disp, win, menuGC, x, y+h-1, w, 1);
    XFillRectangle(disp, win, menuGC, x+w-1, y, 1, h);
}

static void DrawTrianglePattern(Window win, int l, int r, int u,int b)
{
    int m = (u + b)/2;
    XSetForeground(disp, menuGC, menuDimGray);
    XDrawLine(disp, win, menuGC, l, u, l, b);
    XDrawLine(disp, win, menuGC, r, m, l, u);
    XSetForeground(disp, menuGC, menuWhite);
    XDrawLine(disp, win, menuGC, l, b, r, m);
}

static void hiliteMenuItem(Menu *menu, int i)
{
    Window win   = menu->item[i].win;
    char  *label = menu->item[i].label;
    unsigned int width, height;

    XGetWindowSize(disp, win, &width, &height);
    XSetForeground(disp, menuGC, menuLightGray);
    XFillRectangle(disp, win, menuGC, 0, 1, width-1, height-2);
    XSetForeground(disp, menuGC, menuBlack);
    XDrawString(disp, win, menuGC, 7, 14, label, strlen(label));
    if (menu->item[i].submenu)
	DrawTrianglePattern(win, width-15, width-5, 6, 14);
    XFlush(disp);
}

static void paintMenu(Menu *menu, int hilite)
{
    int i;
    unsigned int width, height=ITEM_HEIGHT;

    XGetWindowSize(disp, menu->win, &width, &JunkHeight);
    for (i = 0; i < menu->num_item; i++) {
	Window win   = menu->item[i].win;
	char  *label = menu->item[i].label;
	XSetForeground(disp, menuGC, menuGray);
	XFillRectangle(disp, win, menuGC, 0, 0, width, height);
	RelieveRectangle(win, 0, 0, width, height);
	XSetForeground(disp, menuGC, menuBlack);
	XDrawString(disp, win, menuGC, 7, 14, label, strlen(label));
	if (menu->item[i].submenu)
	    DrawTrianglePattern(win, width-15, width-5, 6, 14);
    }
    if (hilite) {
	int mx, my; 
	XGetMousePos(disp, menu->win, &mx, &my);
	if ((i=my/ITEM_HEIGHT) >= 0 && i < menu->num_item)
	    hiliteMenuItem(menu, i);
    }
    XFlush(disp);
}

static void (*menuCB)(int id);
static int  mapMenu(Menu *menu, int x, int y);

#define NONE  -1

void PopupMenu(Menu *menu, void (*menufunc)(int id)) 
{
    static int init_done = 0;
    int mx, my;
    int select;

    menuCB = menufunc;
    if (!init_done) { InitMenu(menu); init_done = 1; }
    XGetMousePos(disp, rwin, &mx, &my);
    if ((select = mapMenu(menu, mx+1, my)) != NONE) {
	menuCB(select);
    }
}


static void grabPointer(Window win) 
{
    XGrabPointer(disp, win, True, ButtonPressMask|ButtonReleaseMask, 
		 GrabModeAsync, GrabModeAsync, rwin, menuCursor, CurrentTime);
    XFlush(disp);
}

static int mapMenu(Menu *menu, int x, int y)
{
    unsigned int width, height;
    unsigned int sx, sy; /* upper-left of submenu->win */
    int mx, my; /* Mouse_Pos  in menu->win */

    XGetWindowSize(disp, menu->win, &width, &height);
    XMoveWindow(disp, menu->win, x, y);
    XMapRaised(disp, menu->win);
    paintMenu(menu, 1);
    grabPointer(menu->win);

    for (; ;) {
	static int index = NONE;
	static XEvent event;
	int id  = 0;

	XNextEvent(disp, &event);
	switch (event.type) {
	case ButtonRelease:
	    if (index >= 0 && index < menu->num_item)
		if (menu->item[index].submenu) {
		    sx = x + width + MENU_BORDER;
		    sy = y + ITEM_HEIGHT*(index);
		    XWarpPointer(disp, 0, rwin, 0, 0, 0, 0, sx, sy+10);
		    if (id = mapMenu(menu->item[index].submenu, sx, sy)) {
			XUnmapWindow(disp, menu->win);
			return id;
		    }
		    paintMenu(menu, 1);
		    break;
		} else {
		    id = menu->item[index].id;
		    XUnmapWindow(disp, menu->win);
		    return id;
		}
	    XUnmapWindow(disp, menu->win);
	    return (id=index=NONE);
	case LeaveNotify:
	    index = NONE;
	    paintMenu(menu, 0);
	    break;
	case EnterNotify:
	    XGetMousePos(disp, menu->win, &mx, &my);
	    if ((index = my/ITEM_HEIGHT) >= 0 && index < menu->num_item)
		hiliteMenuItem(menu, index);
	    break;
	case MotionNotify:
	    XGetMousePos(disp, menu->win, &mx, &my);
	    if ((index = my/ITEM_HEIGHT) >= 0 && index < menu->num_item)
		if (menu->item[index].submenu && mx > (int)width - 15) {
		    sx = x + width + MENU_BORDER;
		    sy = y + ITEM_HEIGHT*(index);
		    if (id = mapMenu(menu->item[index].submenu, sx, sy)) {
			XUnmapWindow(disp, menu->win);
			return id;
		    }
		    paintMenu(menu, 1);
		    break;
		}
	    if (menu->parent && mx < -15) {	/* leave submenu */
		grabPointer(menu->parent->win);
		XUnmapWindow(disp, menu->win);
		return 0;
	    }
	default:
	    break;
	}
    }
}

#undef NONE
/** end of smenu_x.c **/
